from vocode.turn_based.transcriber.base_transcriber import BaseTranscriber
from vocode.turn_based.transcriber.sr_transcriber import SpeechRecognitionTranscriber
from vocode.turn_based.transcriber.whisper_cpp_transcriber import WhisperCPPTranscriber
from vocode.turn_based.transcriber.whisper_transcriber import WhisperTranscriber
